# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------

import unittest

from azure.mgmt.core.tools import (
    parse_resource_id,
    is_valid_resource_id,
    resource_id,
    is_valid_resource_name,
)


class TestTools(unittest.TestCase):
    def test_resource_parse(self):
        """
        Tests resource id parsing, reforming, and validation.
        """
        tests = [
            {
                "resource_id": "/subscriptions/fakesub/resourcegroups/testgroup/providers"
                "/Microsoft.Storage/storageAccounts/foo/providers"
                "/Microsoft.Authorization/locks/bar",
                "expected": {
                    "name": "foo",
                    "type": "storageAccounts",
                    "namespace": "Microsoft.Storage",
                    "child_name_1": "bar",
                    "child_namespace_1": "Microsoft.Authorization",
                    "child_type_1": "locks",
                    "child_parent_1": "storageAccounts/foo/providers/Microsoft.Authorization/",
                    "resource_group": "testgroup",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/subscriptions/fakesub/resourcegroups/testgroup/providers"
                "/Microsoft.Storage/storageAccounts/foo"
                "/locks/bar",
                "expected": {
                    "name": "foo",
                    "type": "storageAccounts",
                    "namespace": "Microsoft.Storage",
                    "child_name_1": "bar",
                    "child_type_1": "locks",
                    "child_parent_1": "storageAccounts/foo/",
                    "resource_group": "testgroup",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/subscriptions/fakesub/resourcegroups/testgroup/providers"
                "/Microsoft.Storage/storageAccounts/foo/providers"
                "/Microsoft.Authorization/locks/bar/providers/Microsoft.Network/"
                "nets/gc",
                "expected": {
                    "name": "foo",
                    "type": "storageAccounts",
                    "namespace": "Microsoft.Storage",
                    "child_name_1": "bar",
                    "child_namespace_1": "Microsoft.Authorization",
                    "child_type_1": "locks",
                    "child_parent_1": "storageAccounts/foo/providers/Microsoft.Authorization/",
                    "child_name_2": "gc",
                    "child_namespace_2": "Microsoft.Network",
                    "child_type_2": "nets",
                    "child_parent_2": "storageAccounts/foo/providers/Microsoft.Authorization/"
                    "locks/bar/providers/Microsoft.Network/",
                    "resource_group": "testgroup",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/subscriptions/fakesub/resourcegroups/testgroup/providers"
                "/Microsoft.Storage/storageAccounts/foo"
                "/locks/bar/nets/gc",
                "expected": {
                    "name": "foo",
                    "type": "storageAccounts",
                    "namespace": "Microsoft.Storage",
                    "child_name_1": "bar",
                    "child_type_1": "locks",
                    "child_parent_1": "storageAccounts/foo/",
                    "child_name_2": "gc",
                    "child_type_2": "nets",
                    "child_parent_2": "storageAccounts/foo/locks/bar/",
                    "resource_group": "testgroup",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg/providers/"
                "Microsoft.Provider1/resourceType1/name1",
                "expected": {
                    "subscription": "mySub",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.Provider1",
                    "type": "resourceType1",
                    "name": "name1",
                    "resource_parent": "",
                    "resource_namespace": "Microsoft.Provider1",
                    "resource_type": "resourceType1",
                    "resource_name": "name1",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg/providers/"
                "Microsoft.Provider1/resourceType1/name1/resourceType2/name2",
                "expected": {
                    "subscription": "mySub",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.Provider1",
                    "type": "resourceType1",
                    "name": "name1",
                    "child_namespace_1": None,
                    "child_type_1": "resourceType2",
                    "child_name_1": "name2",
                    "child_parent_1": "resourceType1/name1/",
                    "resource_parent": "resourceType1/name1/",
                    "resource_namespace": "Microsoft.Provider1",
                    "resource_type": "resourceType2",
                    "resource_name": "name2",
                },
            },
            {
                "resource_id": "/subscriptions/00000/resourceGroups/myRg/providers/"
                "Microsoft.RecoveryServices/vaults/vault_name/backupFabrics/"
                "fabric_name/protectionContainers/container_name/"
                "protectedItems/item_name/recoveryPoint/recovery_point_guid",
                "expected": {
                    "subscription": "00000",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.RecoveryServices",
                    "type": "vaults",
                    "name": "vault_name",
                    "child_type_1": "backupFabrics",
                    "child_name_1": "fabric_name",
                    "child_parent_1": "vaults/vault_name/",
                    "child_type_2": "protectionContainers",
                    "child_name_2": "container_name",
                    "child_parent_2": "vaults/vault_name/backupFabrics/fabric_name/",
                    "child_type_3": "protectedItems",
                    "child_name_3": "item_name",
                    "child_parent_3": "vaults/vault_name/backupFabrics/fabric_name/"
                    "protectionContainers/container_name/",
                    "child_type_4": "recoveryPoint",
                    "child_name_4": "recovery_point_guid",
                    "child_parent_4": "vaults/vault_name/backupFabrics/fabric_name/"
                    "protectionContainers/container_name/protectedItems/"
                    "item_name/",
                    "resource_parent": "vaults/vault_name/backupFabrics/fabric_name/"
                    "protectionContainers/container_name/protectedItems/"
                    "item_name/",
                    "resource_namespace": "Microsoft.RecoveryServices",
                    "resource_type": "recoveryPoint",
                    "resource_name": "recovery_point_guid",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg/providers/"
                "Microsoft.Provider1/resourceType1/name1/resourceType2/name2/"
                "providers/Microsoft.Provider3/resourceType3/name3",
                "expected": {
                    "subscription": "mySub",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.Provider1",
                    "type": "resourceType1",
                    "name": "name1",
                    "child_namespace_1": None,
                    "child_type_1": "resourceType2",
                    "child_name_1": "name2",
                    "child_parent_1": "resourceType1/name1/",
                    "child_namespace_2": "Microsoft.Provider3",
                    "child_type_2": "resourceType3",
                    "child_name_2": "name3",
                    "child_parent_2": "resourceType1/name1/resourceType2/name2/" "providers/Microsoft.Provider3/",
                    "resource_parent": "resourceType1/name1/resourceType2/name2/" "providers/Microsoft.Provider3/",
                    "resource_namespace": "Microsoft.Provider1",
                    "resource_type": "resourceType3",
                    "resource_name": "name3",
                },
            },
            {
                "resource_id": "/subscriptions/fakesub/providers/Microsoft.Authorization" "/locks/foo",
                "expected": {
                    "name": "foo",
                    "type": "locks",
                    "namespace": "Microsoft.Authorization",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/Subscriptions/fakesub/providers/Microsoft.Authorization" "/locks/foo",
                "expected": {
                    "name": "foo",
                    "type": "locks",
                    "namespace": "Microsoft.Authorization",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg",
                "expected": {"subscription": "mySub", "resource_group": "myRg"},
            },
        ]
        for test in tests:
            self.assertTrue(is_valid_resource_id(test["resource_id"]))
            kwargs = parse_resource_id(test["resource_id"])
            for key in test["expected"]:
                try:
                    self.assertEqual(kwargs[key], test["expected"][key])
                except KeyError:
                    self.assertTrue(key not in kwargs and test["expected"][key] is None)

        invalid_ids = [
            "/subscriptions/fakesub/resourceGroups/myRg/type1/name1",
            "/subscriptions/fakesub/resourceGroups/myRg/providers/Microsoft.Provider/foo",
            "/subscriptions/fakesub/resourceGroups/myRg/providers/namespace/type/name/type1",
            "/subscriptions/fakesub/resourceGroups/",
            "/subscriptions//resourceGroups/",
        ]
        for invalid_id in invalid_ids:
            self.assertFalse(is_valid_resource_id(invalid_id))

        tests = [
            {
                "resource_id": "/subscriptions/fakesub/resourcegroups/testgroup/providers"
                "/Microsoft.Storage/storageAccounts/foo/providers"
                "/Microsoft.Authorization/locks/bar",
                "id_args": {
                    "name": "foo",
                    "type": "storageAccounts",
                    "namespace": "Microsoft.Storage",
                    "child_name_1": "bar",
                    "child_namespace_1": "Microsoft.Authorization",
                    "child_type_1": "locks",
                    "resource_group": "testgroup",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/subscriptions/fakesub/resourcegroups/testgroup/providers"
                "/Microsoft.Storage/storageAccounts/foo"
                "/locks/bar",
                "id_args": {
                    "name": "foo",
                    "type": "storageAccounts",
                    "namespace": "Microsoft.Storage",
                    "child_name_1": "bar",
                    "child_type_1": "locks",
                    "resource_group": "testgroup",
                    "subscription": "fakesub",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg/providers/"
                "Microsoft.Provider1/resourceType1/name1/resourceType2/name2/"
                "providers/Microsoft.Provider3/resourceType3/name3",
                "id_args": {
                    "subscription": "mySub",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.Provider1",
                    "type": "resourceType1",
                    "name": "name1",
                    "child_type_1": "resourceType2",
                    "child_name_1": "name2",
                    "child_namespace_2": "Microsoft.Provider3",
                    "child_type_2": "resourceType3",
                    "child_name_2": "name3",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg/" "providers/Microsoft.Provider1",
                "id_args": {
                    "subscription": "mySub",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.Provider1",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg",
                "id_args": {"subscription": "mySub", "resource_group": "myRg"},
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg/"
                "providers/Microsoft.Provider1/resourceType1/name1/resourceType2/"
                "name2/providers/Microsoft.Provider3",
                "id_args": {
                    "subscription": "mySub",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.Provider1",
                    "type": "resourceType1",
                    "name": "name1",
                    "child_type_1": "resourceType2",
                    "child_name_1": "name2",
                    "child_namespace_2": "Microsoft.Provider3",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg/"
                "providers/Microsoft.Provider1/resourceType1/name1",
                "id_args": {
                    "subscription": "mySub",
                    "resource_group": "myRg",
                    "namespace": "Microsoft.Provider1",
                    "type": "resourceType1",
                    "name": "name1",
                    "child_type_1": None,
                    "child_name_1": "name2",
                    "child_namespace_2": "Microsoft.Provider3",
                },
            },
            {
                "resource_id": "/subscriptions/mySub/resourceGroups/myRg",
                "id_args": {"subscription": "mySub", "resource_group": "myRg"},
            },
        ]
        for test in tests:
            rsrc_id = resource_id(**test["id_args"])
            self.assertEqual(rsrc_id.lower(), test["resource_id"].lower())

    def test_is_resource_name(self):
        invalid_names = [
            "",
            "knights/ni",
            "spam&eggs",
            "i<3you",
            "a" * 261,
        ]

        for test in invalid_names:
            assert not is_valid_resource_name(test)

        valid_names = [
            "abc-123",
            " ",  # no one said it had to be a good resource name.
            "a" * 260,
        ]

        for test in valid_names:
            assert is_valid_resource_name(test)


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
from typing import Optional, Dict, Any, Sequence

from azure.core.polling.base_polling import LocationPolling, StatusCheckPolling, LongRunningOperation
from azure.core.polling.async_base_polling import AsyncLROBasePolling

from .arm_polling import AzureAsyncOperationPolling, BodyContentPolling, HttpRequestTypeVar, AllHttpResponseTypeVar


class AsyncARMPolling(AsyncLROBasePolling):
    def __init__(
        self,
        timeout: float = 30,
        lro_algorithms: Optional[Sequence[LongRunningOperation[HttpRequestTypeVar, AllHttpResponseTypeVar]]] = None,
        lro_options: Optional[Dict[str, Any]] = None,
        path_format_arguments: Optional[Dict[str, str]] = None,
        **operation_config: Any
    ) -> None:
        lro_algorithms = lro_algorithms or [
            AzureAsyncOperationPolling(lro_options=lro_options),
            LocationPolling(),
            BodyContentPolling(),
            StatusCheckPolling(),
        ]
        super(AsyncLROBasePolling, self).__init__(
            timeout=timeout,
            lro_algorithms=lro_algorithms,
            lro_options=lro_options,
            path_format_arguments=path_format_arguments,
            **operation_config
        )


__all__ = ["AsyncARMPolling"]



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import base64
import time
from unittest.mock import Mock

from azure.core.credentials import AccessToken
from azure.core.pipeline import AsyncPipeline
from azure.mgmt.core.policies import (
    AsyncARMChallengeAuthenticationPolicy,
    AsyncAuxiliaryAuthenticationPolicy,
)
from azure.core.pipeline.transport import HttpRequest

import pytest

pytestmark = pytest.mark.asyncio


async def test_claims_challenge():
    """AsyncAsyncARMChallengeAuthenticationPolicy should pass claims from an authentication challenge to its credential"""

    first_token = AccessToken("first", int(time.time()) + 3600)
    second_token = AccessToken("second", int(time.time()) + 3600)
    tokens = (t for t in (first_token, second_token))

    expected_claims = '{"access_token": {"essential": "true"}'
    expected_scope = "scope"

    challenge = 'Bearer authorization_uri="https://localhost", error=".", error_description=".", claims="{}"'.format(
        base64.b64encode(expected_claims.encode()).decode()
    )
    responses = (
        r
        for r in (
            Mock(status_code=401, headers={"WWW-Authenticate": challenge}),
            Mock(status_code=200),
        )
    )

    async def send(request):
        res = next(responses)
        if res.status_code == 401:
            expected_token = first_token.token
        else:
            expected_token = second_token.token
        assert request.headers["Authorization"] == "Bearer " + expected_token

        return res

    async def get_token(*scopes, **kwargs):
        assert scopes == (expected_scope,)
        return next(tokens)

    credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
    transport = Mock(send=Mock(wraps=send))
    policies = [AsyncARMChallengeAuthenticationPolicy(credential, expected_scope)]
    pipeline = AsyncPipeline(transport=transport, policies=policies)

    response = await pipeline.run(HttpRequest("GET", "https://localhost"))

    assert response.http_response.status_code == 200
    assert transport.send.call_count == 2
    assert credential.get_token.call_count == 2

    args, kwargs = credential.get_token.call_args
    assert expected_scope in args
    assert kwargs["claims"] == expected_claims

    with pytest.raises(StopIteration):
        next(tokens)
    with pytest.raises(StopIteration):
        next(responses)


async def test_multiple_claims_challenges():
    """ARMChallengeAuthenticationPolicy should not attempt to handle a response having multiple claims challenges"""

    expected_header = ",".join(
        (
            'Bearer realm="", authorization_uri="https://login.microsoftonline.com/common/oauth2/authorize", client_id="00000003-0000-0000-c000-000000000000", error="insufficient_claims", claims="eyJhY2Nlc3NfdG9rZW4iOiB7ImZvbyI6ICJiYXIifX0="',
            'Bearer authorization_uri="https://login.windows-ppe.net/", error="invalid_token", error_description="User session has been revoked", claims="eyJhY2Nlc3NfdG9rZW4iOnsibmJmIjp7ImVzc2VudGlhbCI6dHJ1ZSwgInZhbHVlIjoiMTYwMzc0MjgwMCJ9fX0="',
        )
    )

    async def send(request):
        return Mock(status_code=401, headers={"WWW-Authenticate": expected_header})

    async def get_token(*_, **__):
        return AccessToken("***", 42)

    transport = Mock(send=Mock(wraps=send))
    credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
    policies = [AsyncARMChallengeAuthenticationPolicy(credential, "scope")]
    pipeline = AsyncPipeline(transport=transport, policies=policies)

    response = await pipeline.run(HttpRequest("GET", "https://localhost"))

    assert transport.send.call_count == 1
    assert credential.get_token.call_count == 1

    # the policy should have returned the error response because it was unable to handle the challenge
    assert response.http_response.status_code == 401
    assert response.http_response.headers["WWW-Authenticate"] == expected_header


async def test_auxiliary_authentication_policy():
    """The auxiliary authentication policy should add a header containing a token from its credential"""
    first_token = AccessToken("first", int(time.time()) + 3600)
    second_token = AccessToken("second", int(time.time()) + 3600)

    async def verify_authorization_header(request):
        assert request.http_request.headers["x-ms-authorization-auxiliary"] == ", ".join(
            "Bearer {}".format(token.token) for token in [first_token, second_token]
        )
        return Mock()

    get_token_calls1 = 0
    get_token_calls2 = 0

    async def get_token1(_):
        nonlocal get_token_calls1
        get_token_calls1 += 1
        return first_token

    async def get_token2(_):
        nonlocal get_token_calls2
        get_token_calls2 += 1
        return second_token

    fake_credential1 = Mock(spec_set=["get_token"], get_token=get_token1)
    fake_credential2 = Mock(spec_set=["get_token"], get_token=get_token2)
    policies = [
        AsyncAuxiliaryAuthenticationPolicy([fake_credential1, fake_credential2], "scope"),
        Mock(send=verify_authorization_header),
    ]

    pipeline = AsyncPipeline(transport=Mock(), policies=policies)
    await pipeline.run(HttpRequest("GET", "https://spam.eggs"))

    assert get_token_calls1 == 1
    assert get_token_calls2 == 1

    await pipeline.run(HttpRequest("GET", "https://spam.eggs"))

    # Didn't need a new token
    assert get_token_calls1 == 1
    assert get_token_calls2 == 1



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
import asyncio
import json
import logging
import uuid
from typing import Union

from azure.core.pipeline import PipelineRequest, PipelineResponse
from azure.core.pipeline.policies import AsyncHTTPPolicy
from azure.core.pipeline.transport import (
    HttpRequest as LegacyHttpRequest,
    AsyncHttpResponse as LegacyAsyncHttpResponse,
)
from azure.core.rest import HttpRequest, AsyncHttpResponse


from ._base import _SansIOARMAutoResourceProviderRegistrationPolicy

_LOGGER = logging.getLogger(__name__)

HTTPRequestType = Union[LegacyHttpRequest, HttpRequest]
AsyncHTTPResponseType = Union[LegacyAsyncHttpResponse, AsyncHttpResponse]
PipelineResponseType = PipelineResponse[HTTPRequestType, AsyncHTTPResponseType]


class AsyncARMAutoResourceProviderRegistrationPolicy(
    _SansIOARMAutoResourceProviderRegistrationPolicy, AsyncHTTPPolicy[HTTPRequestType, AsyncHTTPResponseType]
):  # pylint: disable=name-too-long
    """Auto register an ARM resource provider if not done yet."""

    async def send(  # pylint: disable=invalid-overridden-method
        self, request: PipelineRequest[HTTPRequestType]
    ) -> PipelineResponse[HTTPRequestType, AsyncHTTPResponseType]:
        http_request = request.http_request
        response = await self.next.send(request)
        if response.http_response.status_code == 409:
            rp_name = self._check_rp_not_registered_err(response)
            if rp_name:
                url_prefix = self._extract_subscription_url(http_request.url)
                register_rp_status = await self._async_register_rp(request, url_prefix, rp_name)
                if not register_rp_status:
                    return response
                # Change the 'x-ms-client-request-id' otherwise the Azure endpoint
                # just returns the same 409 payload without looking at the actual query
                if "x-ms-client-request-id" in http_request.headers:
                    http_request.headers["x-ms-client-request-id"] = str(uuid.uuid4())
                response = await self.next.send(request)
        return response

    async def _async_register_rp(
        self, initial_request: PipelineRequest[HTTPRequestType], url_prefix: str, rp_name: str
    ) -> bool:
        """Synchronously register the RP is paremeter.

        Return False if we have a reason to believe this didn't work

        :param initial_request: The initial request
        :type initial_request: ~azure.core.pipeline.PipelineRequest
        :param str url_prefix: The url prefix
        :param str rp_name: The resource provider name
        :return: Return False if we have a reason to believe this didn't work
        :rtype: bool
        """
        post_url = "{}providers/{}/register?api-version=2016-02-01".format(url_prefix, rp_name)
        get_url = "{}providers/{}?api-version=2016-02-01".format(url_prefix, rp_name)
        _LOGGER.warning(
            "Resource provider '%s' used by this operation is not registered. We are registering for you.",
            rp_name,
        )
        post_response = await self.next.send(self._build_next_request(initial_request, "POST", post_url))
        if post_response.http_response.status_code != 200:
            _LOGGER.warning("Registration failed. Please register manually.")
            return False

        while True:
            await asyncio.sleep(10)
            get_response = await self.next.send(self._build_next_request(initial_request, "GET", get_url))
            rp_info = json.loads(get_response.http_response.text())
            if rp_info["registrationState"] == "Registered":
                _LOGGER.warning("Registration succeeded.")
                return True



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
from typing import Mapping, MutableMapping, Optional, Type, Union, cast, Dict, Any
import re
import logging
from azure.core import AzureClouds


_LOGGER = logging.getLogger(__name__)
_ARMID_RE = re.compile(
    "(?i)/subscriptions/(?P<subscription>[^/]+)(/resourceGroups/(?P<resource_group>[^/]+))?"
    + "(/providers/(?P<namespace>[^/]+)/(?P<type>[^/]*)/(?P<name>[^/]+)(?P<children>.*))?"
)

_CHILDREN_RE = re.compile(
    "(?i)(/providers/(?P<child_namespace>[^/]+))?/" + "(?P<child_type>[^/]*)/(?P<child_name>[^/]+)"
)

_ARMNAME_RE = re.compile("^[^<>%&:\\?/]{1,260}$")


__all__ = [
    "parse_resource_id",
    "resource_id",
    "is_valid_resource_id",
    "is_valid_resource_name",
    "get_arm_endpoints",
]


def parse_resource_id(rid: str) -> Mapping[str, Union[str, int]]:
    """Parses a resource_id into its various parts.

    Returns a dictionary with a single key-value pair, 'name': rid, if invalid resource id.

    :param rid: The resource id being parsed
    :type rid: str
    :returns: A dictionary with with following key/value pairs (if found):

        - subscription:            Subscription id
        - resource_group:          Name of resource group
        - namespace:               Namespace for the resource provider (i.e. Microsoft.Compute)
        - type:                    Type of the root resource (i.e. virtualMachines)
        - name:                    Name of the root resource
        - child_namespace_{level}: Namespace for the child resource of that level
        - child_type_{level}:      Type of the child resource of that level
        - child_name_{level}:      Name of the child resource of that level
        - last_child_num:          Level of the last child
        - resource_parent:         Computed parent in the following pattern: providers/{namespace}\
        /{parent}/{type}/{name}
        - resource_namespace:      Same as namespace. Note that this may be different than the \
        target resource's namespace.
        - resource_type:           Type of the target resource (not the parent)
        - resource_name:           Name of the target resource (not the parent)

    :rtype: dict[str,str]
    """
    if not rid:
        return {}
    match = _ARMID_RE.match(rid)
    if match:
        result: MutableMapping[str, Union[None, str, int]] = match.groupdict()
        children = _CHILDREN_RE.finditer(cast(Optional[str], result["children"]) or "")
        count = None
        for count, child in enumerate(children):
            result.update({key + "_%d" % (count + 1): group for key, group in child.groupdict().items()})
        result["last_child_num"] = count + 1 if isinstance(count, int) else None
        final_result = _populate_alternate_kwargs(result)
    else:
        final_result = result = {"name": rid}
    return {key: value for key, value in final_result.items() if value is not None}


def _populate_alternate_kwargs(
    kwargs: MutableMapping[str, Union[None, str, int]]
) -> Mapping[str, Union[None, str, int]]:
    """Translates the parsed arguments into a format used by generic ARM commands
    such as the resource and lock commands.

    :param any kwargs: The parsed arguments
    :return: The translated arguments
    :rtype: any
    """

    resource_namespace = kwargs["namespace"]
    resource_type = kwargs.get("child_type_{}".format(kwargs["last_child_num"])) or kwargs["type"]
    resource_name = kwargs.get("child_name_{}".format(kwargs["last_child_num"])) or kwargs["name"]

    _get_parents_from_parts(kwargs)
    kwargs["resource_namespace"] = resource_namespace
    kwargs["resource_type"] = resource_type
    kwargs["resource_name"] = resource_name
    return kwargs


def _get_parents_from_parts(kwargs: MutableMapping[str, Union[None, str, int]]) -> Mapping[str, Union[None, str, int]]:
    """Get the parents given all the children parameters.

    :param any kwargs: The children parameters
    :return: The parents
    :rtype: any
    """
    parent_builder = []
    if kwargs["last_child_num"] is not None:
        parent_builder.append("{type}/{name}/".format(**kwargs))
        for index in range(1, cast(int, kwargs["last_child_num"])):
            child_namespace = kwargs.get("child_namespace_{}".format(index))
            if child_namespace is not None:
                parent_builder.append("providers/{}/".format(child_namespace))
            kwargs["child_parent_{}".format(index)] = "".join(parent_builder)
            parent_builder.append("{{child_type_{0}}}/{{child_name_{0}}}/".format(index).format(**kwargs))
        child_namespace = kwargs.get("child_namespace_{}".format(kwargs["last_child_num"]))
        if child_namespace is not None:
            parent_builder.append("providers/{}/".format(child_namespace))
        kwargs["child_parent_{}".format(kwargs["last_child_num"])] = "".join(parent_builder)
    kwargs["resource_parent"] = "".join(parent_builder) if kwargs["name"] else None
    return kwargs


def resource_id(**kwargs: Optional[str]) -> str:  # pylint: disable=docstring-keyword-should-match-keyword-only
    """Create a valid resource id string from the given parts.

    This method builds the resource id from the left until the next required id parameter
    to be appended is not found. It then returns the built up id.

    :keyword str subscription: (required) Subscription id
    :keyword str resource_group: Name of resource group
    :keyword str namespace: Namespace for the resource provider (i.e. Microsoft.Compute)
    :keyword str type: Type of the resource (i.e. virtualMachines)
    :keyword str name: Name of the resource (or parent if child_name is also specified)
    :keyword str child_namespace_{level}: Namespace for the child resource of that level (optional)
    :keyword str child_type_{level}: Type of the child resource of that level
    :keyword str child_name_{level}: Name of the child resource of that level

    :returns: A resource id built from the given arguments.
    :rtype: str
    """
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    rid_builder = ["/subscriptions/{subscription}".format(**kwargs)]
    try:
        try:
            rid_builder.append("resourceGroups/{resource_group}".format(**kwargs))
        except KeyError:
            pass
        rid_builder.append("providers/{namespace}".format(**kwargs))
        rid_builder.append("{type}/{name}".format(**kwargs))
        count = 1
        while True:
            try:
                rid_builder.append("providers/{{child_namespace_{}}}".format(count).format(**kwargs))
            except KeyError:
                pass
            rid_builder.append("{{child_type_{0}}}/{{child_name_{0}}}".format(count).format(**kwargs))
            count += 1
    except KeyError:
        pass
    return "/".join(rid_builder)


def is_valid_resource_id(rid: str, exception_type: Optional[Type[BaseException]] = None) -> bool:
    """Validates the given resource id.

    :param rid: The resource id being validated.
    :type rid: str
    :param exception_type: Raises this Exception if invalid.
    :type exception_type: Exception
    :returns: A boolean describing whether the id is valid.
    :rtype: bool
    """
    is_valid: bool = False
    try:
        # Ideally, we would make a TypedDict here, but keeping this file simple for now.
        is_valid = rid and resource_id(**parse_resource_id(rid)).lower() == rid.lower()  # type: ignore
    except KeyError:
        pass
    if not is_valid and exception_type:
        raise exception_type()
    return is_valid


def is_valid_resource_name(rname: str, exception_type: Optional[Type[BaseException]] = None) -> bool:
    """Validates the given resource name to ARM guidelines, individual services may be more restrictive.

    :param rname: The resource name being validated.
    :type rname: str
    :param exception_type: Raises this Exception if invalid.
    :type exception_type: Exception
    :returns: A boolean describing whether the name is valid.
    :rtype: bool
    """

    match = _ARMNAME_RE.match(rname)

    if match:
        return True
    if exception_type:
        raise exception_type()
    return False


def get_arm_endpoints(cloud_setting: AzureClouds) -> Dict[str, Any]:
    """Get the ARM endpoint and ARM credential scopes for the given cloud setting.

    :param cloud_setting: The cloud setting for which to get the ARM endpoint.
    :type cloud_setting: AzureClouds
    :return: The ARM endpoint and ARM credential scopes.
    :rtype: dict[str, Any]
    """
    if cloud_setting == AzureClouds.AZURE_CHINA_CLOUD:
        return {
            "resource_manager": "https://management.chinacloudapi.cn/",
            "credential_scopes": ["https://management.chinacloudapi.cn/.default"],
        }
    if cloud_setting == AzureClouds.AZURE_US_GOVERNMENT:
        return {
            "resource_manager": "https://management.usgovcloudapi.net/",
            "credential_scopes": ["https://management.core.usgovcloudapi.net/.default"],
        }
    if cloud_setting == AzureClouds.AZURE_PUBLIC_CLOUD:
        return {
            "resource_manager": "https://management.azure.com/",
            "credential_scopes": ["https://management.azure.com/.default"],
        }
    raise ValueError("Unknown cloud setting: {}".format(cloud_setting))



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
from typing import Mapping, Any, Sequence
import json
import logging


from azure.core.exceptions import ODataV4Format


_LOGGER = logging.getLogger(__name__)


class TypedErrorInfo:
    """Additional info class defined in ARM specification.

    https://github.com/Azure/azure-resource-manager-rpc/blob/master/v1.0/common-api-details.md#error-response-content
    """

    def __init__(self, type: str, info: Mapping[str, Any]) -> None:  # pylint: disable=redefined-builtin
        self.type = type
        self.info = info

    def __str__(self) -> str:
        """Cloud error message.

        :return: The cloud error message.
        :rtype: str
        """
        error_str = "Type: {}".format(self.type)
        error_str += "\nInfo: {}".format(json.dumps(self.info, indent=4))
        return error_str


class ARMErrorFormat(ODataV4Format):
    """Describe error format from ARM, used at the base or inside "details" node.

    This format is compatible with ODataV4 format.
    https://github.com/Azure/azure-resource-manager-rpc/blob/master/v1.0/common-api-details.md#error-response-content
    """

    def __init__(self, json_object: Mapping[str, Any]) -> None:
        # Parse the ODatav4 part
        super(ARMErrorFormat, self).__init__(json_object)
        if "error" in json_object:
            json_object = json_object["error"]

        # ARM specific annotations
        self.additional_info: Sequence[TypedErrorInfo] = [
            TypedErrorInfo(additional_info["type"], additional_info["info"])
            for additional_info in json_object.get("additionalInfo", [])
        ]

    def __str__(self) -> str:
        error_str = super(ARMErrorFormat, self).__str__()

        if self.additional_info:
            error_str += "\nAdditional Information:"
            for error_info in self.additional_info:
                error_str += str(error_info)

        return error_str



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------

from azure.mgmt.core import AsyncARMPipelineClient
from azure.mgmt.core.policies import ARMHttpLoggingPolicy
from azure.core.configuration import Configuration


def test_default_http_logging_policy():
    config = Configuration()
    pipeline_client = AsyncARMPipelineClient(base_url="test", config=config)
    http_logging_policy = pipeline_client._pipeline._impl_policies[-1]._policy
    assert http_logging_policy.allowed_header_names == ARMHttpLoggingPolicy.DEFAULT_HEADERS_WHITELIST


def test_pass_in_http_logging_policy():
    config = Configuration()
    http_logging_policy = ARMHttpLoggingPolicy()
    http_logging_policy.allowed_header_names.update({"x-ms-added-header"})
    config.http_logging_policy = http_logging_policy

    pipeline_client = AsyncARMPipelineClient(base_url="test", config=config)
    http_logging_policy = pipeline_client._pipeline._impl_policies[-1]._policy
    assert http_logging_policy.allowed_header_names == ARMHttpLoggingPolicy.DEFAULT_HEADERS_WHITELIST.union(
        {"x-ms-added-header"}
    )



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
from typing import cast, Awaitable, Optional, List, Union, Any
import inspect

from azure.core.pipeline.policies import (
    AsyncBearerTokenCredentialPolicy,
    AsyncHTTPPolicy,
)
from azure.core.pipeline import PipelineRequest, PipelineResponse
from azure.core.pipeline.transport import (
    HttpRequest as LegacyHttpRequest,
    AsyncHttpResponse as LegacyAsyncHttpResponse,
)
from azure.core.rest import HttpRequest, AsyncHttpResponse
from azure.core.credentials import AccessToken
from azure.core.credentials_async import AsyncTokenCredential


from ._authentication import _parse_claims_challenge, _AuxiliaryAuthenticationPolicyBase


HTTPRequestType = Union[LegacyHttpRequest, HttpRequest]
AsyncHTTPResponseType = Union[LegacyAsyncHttpResponse, AsyncHttpResponse]


async def await_result(func, *args, **kwargs):
    """If func returns an awaitable, await it.

    :param callable func: Function to call
    :param any args: Positional arguments to pass to func
    :return: Result of func
    :rtype: any
    """
    result = func(*args, **kwargs)
    if inspect.isawaitable(result):
        return await result
    return result


class AsyncARMChallengeAuthenticationPolicy(AsyncBearerTokenCredentialPolicy):
    """Adds a bearer token Authorization header to requests.

    This policy internally handles Continuous Access Evaluation (CAE) challenges. When it can't complete a challenge,
    it will return the 401 (unauthorized) response from ARM.

    :param ~azure.core.credentials.TokenCredential credential: credential for authorizing requests
    :param str scopes: required authentication scopes
    """

    # pylint:disable=unused-argument
    async def on_challenge(
        self,
        request: PipelineRequest[HTTPRequestType],
        response: PipelineResponse[HTTPRequestType, AsyncHTTPResponseType],
    ) -> bool:
        """Authorize request according to an ARM authentication challenge

        :param ~azure.core.pipeline.PipelineRequest request: the request which elicited an authentication challenge
        :param ~azure.core.pipeline.PipelineResponse response: the resource provider's response
        :returns: a bool indicating whether the policy should send the request
        :rtype: bool
        """
        # Casting, as the code seems to be certain that on_challenge this header will be present
        challenge: str = cast(str, response.http_response.headers.get("WWW-Authenticate"))
        claims = _parse_claims_challenge(challenge)
        if claims:
            await self.authorize_request(request, *self._scopes, claims=claims)
            return True

        return False


class AsyncAuxiliaryAuthenticationPolicy(
    _AuxiliaryAuthenticationPolicyBase[AsyncTokenCredential],
    AsyncHTTPPolicy[HTTPRequestType, AsyncHTTPResponseType],
):
    async def _get_auxiliary_tokens(self, *scopes: str, **kwargs: Any) -> Optional[List[AccessToken]]:
        if self._auxiliary_credentials:
            return [await cred.get_token(*scopes, **kwargs) for cred in self._auxiliary_credentials]
        return None

    async def on_request(self, request: PipelineRequest[HTTPRequestType]) -> None:
        """Called before the policy sends a request.

        The base implementation authorizes the request with an auxiliary authorization token.

        :param ~azure.core.pipeline.PipelineRequest request: the request
        """
        self._enforce_https(request)

        if self._need_new_aux_tokens:
            self._aux_tokens = await self._get_auxiliary_tokens(*self._scopes)

        self._update_headers(request.http_request.headers)

    def on_response(
        self,
        request: PipelineRequest[HTTPRequestType],
        response: PipelineResponse[HTTPRequestType, AsyncHTTPResponseType],
    ) -> Optional[Awaitable[None]]:
        """Executed after the request comes back from the next policy.

        :param request: Request to be modified after returning from the policy.
        :type request: ~azure.core.pipeline.PipelineRequest
        :param response: Pipeline response object
        :type response: ~azure.core.pipeline.PipelineResponse
        """

    def on_exception(self, request: PipelineRequest[HTTPRequestType]) -> None:
        """Executed when an exception is raised while executing the next policy.

        This method is executed inside the exception handler.

        :param request: The Pipeline request object
        :type request: ~azure.core.pipeline.PipelineRequest
        """
        # pylint: disable=unused-argument
        return

    async def send(
        self, request: PipelineRequest[HTTPRequestType]
    ) -> PipelineResponse[HTTPRequestType, AsyncHTTPResponseType]:
        """Authorize request with a bearer token and send it to the next policy

        :param request: The pipeline request object
        :type request: ~azure.core.pipeline.PipelineRequest
        :return: The pipeline response object
        :rtype: ~azure.core.pipeline.PipelineResponse
        """
        await await_result(self.on_request, request)
        try:
            response = await self.next.send(request)
            await await_result(self.on_response, request, response)
        except Exception:  # pylint:disable=broad-except
            handled = await await_result(self.on_exception, request)
            if not handled:
                raise
        return response



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------

import json
import time
from unittest import mock
import pytest
import httpretty

from azure.core.configuration import Configuration
from azure.core.pipeline import Pipeline
from azure.core.pipeline.transport import (
    HttpRequest,
    RequestsTransport,
)

from azure.mgmt.core import ARMPipelineClient
from azure.mgmt.core.policies import (
    ARMAutoResourceProviderRegistrationPolicy,
    ARMHttpLoggingPolicy,
)


@pytest.fixture
def sleepless(monkeypatch):
    def sleep(_):
        pass

    monkeypatch.setattr(time, "sleep", sleep)


@httpretty.activate
@pytest.mark.usefixtures("sleepless")
def test_register_rp_policy():
    """Protocol:
    - We call the provider and get a 409 provider error
    - Now we POST register provider and get "Registering"
    - Now we GET register provider and get "Registered"
    - We call again the first endpoint and this time this succeed
    """

    provider_url = (
        "https://management.azure.com/"
        "subscriptions/12345678-9abc-def0-0000-000000000000/"
        "resourceGroups/clitest.rg000001/"
        "providers/Microsoft.Sql/servers/ygserver123?api-version=2014-04-01"
    )

    provider_error = (
        '{"error":{"code":"MissingSubscriptionRegistration", '
        '"message":"The subscription registration is in \'Unregistered\' state. '
        "The subscription must be registered to use namespace 'Microsoft.Sql'. "
        'See https://aka.ms/rps-not-found for how to register subscriptions."}}'
    )

    provider_success = '{"success": true}'

    httpretty.register_uri(
        httpretty.PUT,
        provider_url,
        responses=[
            httpretty.Response(body=provider_error, status=409),
            httpretty.Response(body=provider_success),
        ],
        content_type="application/json",
    )

    register_post_url = (
        "https://management.azure.com/"
        "subscriptions/12345678-9abc-def0-0000-000000000000/"
        "providers/Microsoft.Sql/register?api-version=2016-02-01"
    )

    register_post_result = {"registrationState": "Registering"}

    register_get_url = (
        "https://management.azure.com/"
        "subscriptions/12345678-9abc-def0-0000-000000000000/"
        "providers/Microsoft.Sql?api-version=2016-02-01"
    )

    register_get_result = {"registrationState": "Registered"}

    httpretty.register_uri(
        httpretty.POST,
        register_post_url,
        body=json.dumps(register_post_result),
        content_type="application/json",
    )

    httpretty.register_uri(
        httpretty.GET,
        register_get_url,
        body=json.dumps(register_get_result),
        content_type="application/json",
    )

    request = HttpRequest("PUT", provider_url)
    policies = [
        ARMAutoResourceProviderRegistrationPolicy(),
    ]
    with Pipeline(RequestsTransport(), policies=policies) as pipeline:
        response = pipeline.run(request)

    assert json.loads(response.http_response.text())["success"]


@httpretty.activate
@pytest.mark.usefixtures("sleepless")
def test_register_failed_policy():
    """Protocol:
    - We call the provider and get a 409 provider error
    - Now we POST register provider and get "Registering"
    - This POST failed
    """

    provider_url = (
        "https://management.azure.com/"
        "subscriptions/12345678-9abc-def0-0000-000000000000/"
        "resourceGroups/clitest.rg000001/"
        "providers/Microsoft.Sql/servers/ygserver123?api-version=2014-04-01"
    )

    provider_error = (
        '{"error":{"code":"MissingSubscriptionRegistration", '
        '"message":"The subscription registration is in \'Unregistered\' state. '
        "The subscription must be registered to use namespace 'Microsoft.Sql'. "
        'See https://aka.ms/rps-not-found for how to register subscriptions."}}'
    )

    provider_success = '{"success": true}'

    httpretty.register_uri(
        httpretty.PUT,
        provider_url,
        responses=[
            httpretty.Response(body=provider_error, status=409),
            httpretty.Response(body=provider_success),
        ],
        content_type="application/json",
    )

    register_post_url = (
        "https://management.azure.com/"
        "subscriptions/12345678-9abc-def0-0000-000000000000/"
        "providers/Microsoft.Sql/register?api-version=2016-02-01"
    )

    httpretty.register_uri(httpretty.POST, register_post_url, status=409, content_type="application/json")

    request = HttpRequest("PUT", provider_url)
    policies = [
        ARMAutoResourceProviderRegistrationPolicy(),
    ]
    with Pipeline(RequestsTransport(), policies=policies) as pipeline:
        response = pipeline.run(request)

    assert response.http_response.status_code == 409


def test_default_http_logging_policy():
    config = Configuration()
    pipeline_client = ARMPipelineClient(base_url="test", config=config)
    http_logging_policy = pipeline_client._pipeline._impl_policies[-1]._policy
    assert http_logging_policy.allowed_header_names == ARMHttpLoggingPolicy.DEFAULT_HEADERS_WHITELIST
    assert http_logging_policy.allowed_header_names == ARMHttpLoggingPolicy.DEFAULT_HEADERS_ALLOWLIST


def test_pass_in_http_logging_policy():
    config = Configuration()
    http_logging_policy = ARMHttpLoggingPolicy()
    http_logging_policy.allowed_header_names.update({"x-ms-added-header"})
    config.http_logging_policy = http_logging_policy

    pipeline_client = ARMPipelineClient(base_url="test", config=config)
    http_logging_policy = pipeline_client._pipeline._impl_policies[-1]._policy
    assert http_logging_policy.allowed_header_names == ARMHttpLoggingPolicy.DEFAULT_HEADERS_ALLOWLIST.union(
        {"x-ms-added-header"}
    )
    assert http_logging_policy.allowed_header_names == ARMHttpLoggingPolicy.DEFAULT_HEADERS_WHITELIST.union(
        {"x-ms-added-header"}
    )



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import base64
import json
import pickle
import re
import types
from unittest import mock
import pytest

from requests import Request, Response

from azure.core.polling import async_poller
from azure.core.exceptions import DecodeError, HttpResponseError
from azure.core import AsyncPipelineClient
from azure.core.pipeline import PipelineResponse, AsyncPipeline
from azure.core.pipeline.transport import (
    AsyncioRequestsTransportResponse,
    AsyncHttpTransport,
)

from azure.core.polling.base_polling import (
    LongRunningOperation,
    BadStatus,
    LocationPolling,
)
from azure.mgmt.core.polling.async_arm_polling import (
    AsyncARMPolling,
)


class SimpleResource:
    """An implementation of Python 3 SimpleNamespace.
    Used to deserialize resource objects from response bodies where
    no particular object type has been specified.
    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        keys = sorted(self.__dict__)
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class BadEndpointError(Exception):
    pass


TEST_NAME = "foo"
RESPONSE_BODY = {"properties": {"provisioningState": "InProgress"}}
ASYNC_BODY = json.dumps({"status": "Succeeded"})
ASYNC_URL = "http://dummyurlFromAzureAsyncOPHeader_Return200"
LOCATION_BODY = json.dumps({"name": TEST_NAME})
LOCATION_URL = "http://dummyurlurlFromLocationHeader_Return200"
RESOURCE_BODY = json.dumps({"name": TEST_NAME})
RESOURCE_URL = "http://subscriptions/sub1/resourcegroups/g1/resourcetype1/resource1"
ERROR = "http://dummyurl_ReturnError"
POLLING_STATUS = 200

CLIENT = AsyncPipelineClient("http://example.org")


async def mock_run(client_self, request, **kwargs):
    return TestArmPolling.mock_update(request.url)


CLIENT._pipeline.run = types.MethodType(mock_run, CLIENT)


@pytest.fixture
def async_pipeline_client_builder():
    """Build a client that use the "send" callback as final transport layer

    send will receive "request" and kwargs as any transport layer
    """

    def create_client(send_cb):
        class TestHttpTransport(AsyncHttpTransport):
            async def open(self):
                pass

            async def close(self):
                pass

            async def __aexit__(self, *args, **kwargs):
                pass

            async def send(self, request, **kwargs):
                return await send_cb(request, **kwargs)

        return AsyncPipelineClient("http://example.org/", pipeline=AsyncPipeline(transport=TestHttpTransport()))

    return create_client


@pytest.fixture
def deserialization_cb():
    def cb(pipeline_response):
        return json.loads(pipeline_response.http_response.text())

    return cb


@pytest.mark.asyncio
async def test_post(async_pipeline_client_builder, deserialization_cb):

    # Test POST LRO with both Location and Azure-AsyncOperation

    # The initial response contains both Location and Azure-AsyncOperation, a 202 and no Body
    initial_response = TestArmPolling.mock_send(
        "POST",
        202,
        {
            "location": "http://example.org/location",
            "azure-asyncoperation": "http://example.org/async_monitor",
        },
        "",
    )

    async def send(request, **kwargs):
        assert request.method == "GET"

        if request.url == "http://example.org/location":
            return TestArmPolling.mock_send("GET", 200, body={"location_result": True}).http_response
        elif request.url == "http://example.org/async_monitor":
            return TestArmPolling.mock_send("GET", 200, body={"status": "Succeeded"}).http_response
        else:
            pytest.fail("No other query allowed")

    client = async_pipeline_client_builder(send)

    # Test 1, LRO options with Location final state
    poll = async_poller(
        client,
        initial_response,
        deserialization_cb,
        AsyncARMPolling(0, lro_options={"final-state-via": "location"}),
    )
    result = await poll
    assert result["location_result"] == True

    # Test 2, LRO options with Azure-AsyncOperation final state
    poll = async_poller(
        client,
        initial_response,
        deserialization_cb,
        AsyncARMPolling(0, lro_options={"final-state-via": "azure-async-operation"}),
    )
    result = await poll
    assert result["status"] == "Succeeded"

    # Test 3, "do the right thing" and use Location by default
    poll = async_poller(client, initial_response, deserialization_cb, AsyncARMPolling(0))
    result = await poll
    assert result["location_result"] == True

    # Test 4, location has no body

    async def send(request, **kwargs):
        assert request.method == "GET"

        if request.url == "http://example.org/location":
            return TestArmPolling.mock_send("GET", 200, body=None).http_response
        elif request.url == "http://example.org/async_monitor":
            return TestArmPolling.mock_send("GET", 200, body={"status": "Succeeded"}).http_response
        else:
            pytest.fail("No other query allowed")

    client = async_pipeline_client_builder(send)

    poll = async_poller(
        client,
        initial_response,
        deserialization_cb,
        AsyncARMPolling(0, lro_options={"final-state-via": "location"}),
    )
    result = await poll
    assert result is None


class TestArmPolling(object):

    convert = re.compile("([a-z0-9])([A-Z])")

    @staticmethod
    def mock_send(method, status, headers=None, body=RESPONSE_BODY):
        if headers is None:
            headers = {}
        response = Response()
        response._content_consumed = True
        response._content = json.dumps(body).encode("ascii") if body is not None else None
        response.request = Request()
        response.request.method = method
        response.request.url = RESOURCE_URL
        response.request.headers = {"x-ms-client-request-id": "67f4dd4e-6262-45e1-8bed-5c45cf23b6d9"}
        response.status_code = status
        response.headers = headers
        response.headers.update({"content-type": "application/json; charset=utf8"})
        response.reason = "OK"

        request = CLIENT._request(
            response.request.method,
            response.request.url,
            None,  # params
            response.request.headers,
            body,
            None,  # form_content
            None,  # stream_content
        )

        return PipelineResponse(
            request,
            AsyncioRequestsTransportResponse(
                request,
                response,
            ),
            None,  # context
        )

    @staticmethod
    def mock_update(url, headers=None):
        response = Response()
        response._content_consumed = True
        response.request = mock.create_autospec(Request)
        response.request.method = "GET"
        response.headers = headers or {}
        response.headers.update({"content-type": "application/json; charset=utf8"})
        response.reason = "OK"

        if url == ASYNC_URL:
            response.request.url = url
            response.status_code = POLLING_STATUS
            response._content = ASYNC_BODY.encode("ascii")
            response.randomFieldFromPollAsyncOpHeader = None

        elif url == LOCATION_URL:
            response.request.url = url
            response.status_code = POLLING_STATUS
            response._content = LOCATION_BODY.encode("ascii")
            response.randomFieldFromPollLocationHeader = None

        elif url == ERROR:
            raise BadEndpointError("boom")

        elif url == RESOURCE_URL:
            response.request.url = url
            response.status_code = POLLING_STATUS
            response._content = RESOURCE_BODY.encode("ascii")

        else:
            raise Exception("URL does not match")

        request = CLIENT._request(
            response.request.method,
            response.request.url,
            None,  # params
            {},  # request has no headers
            None,  # Request has no body
            None,  # form_content
            None,  # stream_content
        )

        return PipelineResponse(
            request,
            AsyncioRequestsTransportResponse(
                request,
                response,
            ),
            None,  # context
        )

    @staticmethod
    def mock_outputs(pipeline_response):
        response = pipeline_response.http_response
        try:
            body = json.loads(response.text())
        except ValueError:
            raise DecodeError("Impossible to deserialize")

        body = {TestArmPolling.convert.sub(r"\1_\2", k).lower(): v for k, v in body.items()}
        properties = body.setdefault("properties", {})
        if "name" in body:
            properties["name"] = body["name"]
        if properties:
            properties = {TestArmPolling.convert.sub(r"\1_\2", k).lower(): v for k, v in properties.items()}
            del body["properties"]
            body.update(properties)
            resource = SimpleResource(**body)
        else:
            raise DecodeError("Impossible to deserialize")
            resource = SimpleResource(**body)
        return resource

    @staticmethod
    def mock_deserialization_no_body(pipeline_response):
        """Use this mock when you don't expect a return (last body irrelevant)"""
        return None


@pytest.mark.asyncio
async def test_long_running_put():
    # TODO: Test custom header field

    # Test throw on non LRO related status code
    response = TestArmPolling.mock_send("PUT", 1000, {})
    with pytest.raises(HttpResponseError):
        await async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))

    # Test with no polling necessary
    response_body = {
        "properties": {"provisioningState": "Succeeded"},
        "name": TEST_NAME,
    }
    response = TestArmPolling.mock_send("PUT", 201, {}, response_body)

    def no_update_allowed(url, headers=None):
        raise ValueError("Should not try to update")

    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert not hasattr(polling_method._pipeline_response, "randomFieldFromPollAsyncOpHeader")

    # Test polling from azure-asyncoperation header
    response = TestArmPolling.mock_send("PUT", 201, {"azure-asyncoperation": ASYNC_URL})
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert not hasattr(polling_method._pipeline_response, "randomFieldFromPollAsyncOpHeader")

    # Test polling location header
    response = TestArmPolling.mock_send("PUT", 201, {"location": LOCATION_URL})
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollLocationHeader is None

    # Test polling initial payload invalid (SQLDb)
    response_body = {}  # Empty will raise
    response = TestArmPolling.mock_send("PUT", 201, {"location": LOCATION_URL}, response_body)
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollLocationHeader is None

    # Test fail to poll from azure-asyncoperation header
    response = TestArmPolling.mock_send("PUT", 201, {"azure-asyncoperation": ERROR})
    with pytest.raises(BadEndpointError):
        poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))

    # Test fail to poll from location header
    response = TestArmPolling.mock_send("PUT", 201, {"location": ERROR})
    with pytest.raises(BadEndpointError):
        poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))


@pytest.mark.asyncio
async def test_long_running_patch():

    # Test polling from location header
    response = TestArmPolling.mock_send(
        "PATCH",
        202,
        {"location": LOCATION_URL},
        body={"properties": {"provisioningState": "Succeeded"}},
    )
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollLocationHeader is None

    # Test polling from azure-asyncoperation header
    response = TestArmPolling.mock_send(
        "PATCH",
        202,
        {"azure-asyncoperation": ASYNC_URL},
        body={"properties": {"provisioningState": "Succeeded"}},
    )
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert not hasattr(polling_method._pipeline_response, "randomFieldFromPollAsyncOpHeader")

    # Test polling from location header
    response = TestArmPolling.mock_send(
        "PATCH",
        200,
        {"location": LOCATION_URL},
        body={"properties": {"provisioningState": "Succeeded"}},
    )
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollLocationHeader is None

    # Test polling from azure-asyncoperation header
    response = TestArmPolling.mock_send(
        "PATCH",
        200,
        {"azure-asyncoperation": ASYNC_URL},
        body={"properties": {"provisioningState": "Succeeded"}},
    )
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert not hasattr(polling_method._pipeline_response, "randomFieldFromPollAsyncOpHeader")

    # Test fail to poll from azure-asyncoperation header
    response = TestArmPolling.mock_send("PATCH", 202, {"azure-asyncoperation": ERROR})
    with pytest.raises(BadEndpointError):
        poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))

    # Test fail to poll from location header
    response = TestArmPolling.mock_send("PATCH", 202, {"location": ERROR})
    with pytest.raises(BadEndpointError):
        poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))


@pytest.mark.asyncio
async def test_long_running_delete():
    # Test polling from azure-asyncoperation header
    response = TestArmPolling.mock_send("DELETE", 202, {"azure-asyncoperation": ASYNC_URL}, body="")
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_deserialization_no_body, polling_method)
    assert poll is None
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollAsyncOpHeader is None


@pytest.mark.asyncio
async def test_long_running_post():

    # Test polling from azure-asyncoperation header
    response = TestArmPolling.mock_send(
        "POST",
        201,
        {"azure-asyncoperation": ASYNC_URL},
        body={"properties": {"provisioningState": "Succeeded"}},
    )
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_deserialization_no_body, polling_method)
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollAsyncOpHeader is None

    # Test polling from azure-asyncoperation header
    response = TestArmPolling.mock_send(
        "POST",
        202,
        {"azure-asyncoperation": ASYNC_URL},
        body={"properties": {"provisioningState": "Succeeded"}},
    )
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_deserialization_no_body, polling_method)
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollAsyncOpHeader is None

    # Test polling from location header
    response = TestArmPolling.mock_send(
        "POST",
        202,
        {"location": LOCATION_URL},
        body={"properties": {"provisioningState": "Succeeded"}},
    )
    polling_method = AsyncARMPolling(0)
    poll = await async_poller(CLIENT, response, TestArmPolling.mock_outputs, polling_method)
    assert poll.name == TEST_NAME
    assert polling_method._pipeline_response.http_response.internal_response.randomFieldFromPollLocationHeader is None

    # Test fail to poll from azure-asyncoperation header
    response = TestArmPolling.mock_send("POST", 202, {"azure-asyncoperation": ERROR})
    with pytest.raises(BadEndpointError):
        await async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))

    # Test fail to poll from location header
    response = TestArmPolling.mock_send("POST", 202, {"location": ERROR})
    with pytest.raises(BadEndpointError):
        await async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))


@pytest.mark.asyncio
async def test_long_running_negative():
    global LOCATION_BODY
    global POLLING_STATUS

    # Test LRO PUT throws for invalid json
    LOCATION_BODY = "{"
    response = TestArmPolling.mock_send("POST", 202, {"location": LOCATION_URL})
    poll = async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))
    with pytest.raises(DecodeError):
        await poll

    LOCATION_BODY = "{'\"}"
    response = TestArmPolling.mock_send("POST", 202, {"location": LOCATION_URL})
    poll = async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))
    with pytest.raises(DecodeError):
        await poll

    LOCATION_BODY = "{"
    POLLING_STATUS = 203
    response = TestArmPolling.mock_send("POST", 202, {"location": LOCATION_URL})
    poll = async_poller(CLIENT, response, TestArmPolling.mock_outputs, AsyncARMPolling(0))
    with pytest.raises(HttpResponseError) as error:  # TODO: Node.js raises on deserialization
        await poll
    assert error.value.continuation_token == base64.b64encode(pickle.dumps(response)).decode("ascii")

    LOCATION_BODY = json.dumps({"name": TEST_NAME})
    POLLING_STATUS = 200


def test_polling_with_path_format_arguments():
    method = AsyncARMPolling(timeout=0, path_format_arguments={"host": "host:3000", "accountName": "local"})
    client = AsyncPipelineClient(base_url="http://{accountName}{host}")

    method._operation = LocationPolling()
    method._operation._location_url = "/results/1"
    method._client = client
    assert "http://localhost:3000/results/1" == method._client.format_url(
        method._operation.get_polling_url(), **method._path_format_arguments
    )



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import base64
import time

from azure.core.credentials import AccessToken
from azure.core.pipeline import Pipeline
from azure.mgmt.core.policies._authentication import (
    _parse_claims_challenge,
    ARMChallengeAuthenticationPolicy,
    AuxiliaryAuthenticationPolicy,
)

from azure.core.pipeline.transport import HttpRequest
from devtools_testutils.fake_credentials import FakeTokenCredential


import pytest
from unittest.mock import Mock


CLAIM_TOKEN = base64.b64encode(b'{"access_token": {"foo": "bar"}}').decode()
CLAIM_NBF = base64.b64encode(b'{"access_token":{"nbf":{"essential":true, "value":"1603742800"}}}').decode()
ip_claim = b'{"access_token":{"nbf":{"essential":true,"value":"1610563006"},"xms_rp_ipaddr":{"value":"1.2.3.4"}}}'
CLAIM_IP = base64.b64encode(ip_claim).decode()[:-2]  # Trim off padding = characters


@pytest.mark.parametrize(
    "challenge,expected_claims",
    (
        # CAE - insufficient claims
        (
            f'Bearer realm="", authorization_uri="https://login.microsoftonline.com/common/oauth2/authorize", client_id="00000003-0000-0000-c000-000000000000", error="insufficient_claims", claims="{CLAIM_TOKEN}"',
            '{"access_token": {"foo": "bar"}}',
        ),
        # CAE - sessions revoked
        (
            f'Bearer authorization_uri="https://login.windows-ppe.net/", error="invalid_token", error_description="User session has been revoked", claims={CLAIM_NBF}',
            '{"access_token":{"nbf":{"essential":true, "value":"1603742800"}}}',
        ),
        # CAE - IP policy
        (
            f'Bearer authorization_uri="https://login.windows.net/", error="invalid_token", error_description="Tenant IP Policy validate failed.", claims={CLAIM_IP}',
            '{"access_token":{"nbf":{"essential":true,"value":"1610563006"},"xms_rp_ipaddr":{"value":"1.2.3.4"}}}',
        ),
        # ARM
        (
            'Bearer authorization_uri="https://login.windows.net/", error="invalid_token", error_description="The authentication failed because of missing \'Authorization\' header."',
            None,
        ),
    ),
)
def test_challenge_parsing(challenge, expected_claims):
    claims = _parse_claims_challenge(challenge)
    assert claims == expected_claims


def test_auxiliary_authentication_policy():
    """The auxiliary authentication policy should add a header containing a token from its credential"""
    first_token = AccessToken("first", int(time.time()) + 3600)
    second_token = AccessToken("second", int(time.time()) + 3600)

    def verify_authorization_header(request):
        assert request.http_request.headers["x-ms-authorization-auxiliary"] == ", ".join(
            "Bearer {}".format(token.token) for token in [first_token, second_token]
        )
        return Mock()

    fake_credential1 = Mock(spec_set=["get_token"], get_token=Mock(return_value=first_token))
    fake_credential2 = Mock(spec_set=["get_token"], get_token=Mock(return_value=second_token))
    policies = [
        AuxiliaryAuthenticationPolicy([fake_credential1, fake_credential2], "scope"),
        Mock(send=verify_authorization_header),
    ]

    pipeline = Pipeline(transport=Mock(), policies=policies)
    pipeline.run(HttpRequest("GET", "https://spam.eggs"))

    assert fake_credential1.get_token.call_count == 1
    assert fake_credential2.get_token.call_count == 1

    pipeline.run(HttpRequest("GET", "https://spam.eggs"))

    # Didn't need a new token
    assert fake_credential1.get_token.call_count == 1
    assert fake_credential2.get_token.call_count == 1


def test_claims_challenge():
    """ARMChallengeAuthenticationPolicy should pass claims from an authentication challenge to its credential"""

    first_token = AccessToken("first", int(time.time()) + 3600)
    second_token = AccessToken("second", int(time.time()) + 3600)
    tokens = (t for t in (first_token, second_token))

    expected_claims = '{"access_token": {"essential": "true"}'
    expected_scope = "scope"

    challenge = 'Bearer authorization_uri="https://localhost", error=".", error_description=".", claims="{}"'.format(
        base64.b64encode(expected_claims.encode()).decode()
    )
    responses = (
        r
        for r in (
            Mock(status_code=401, headers={"WWW-Authenticate": challenge}),
            Mock(status_code=200),
        )
    )

    def send(request):
        res = next(responses)
        if res.status_code == 401:
            expected_token = first_token.token
        else:
            expected_token = second_token.token
        assert request.headers["Authorization"] == "Bearer " + expected_token

        return res

    def get_token(*scopes, **kwargs):
        assert scopes == (expected_scope,)
        return next(tokens)

    credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
    transport = Mock(send=Mock(wraps=send))
    policies = [ARMChallengeAuthenticationPolicy(credential, expected_scope)]
    pipeline = Pipeline(transport=transport, policies=policies)

    response = pipeline.run(HttpRequest("GET", "https://localhost"))

    assert response.http_response.status_code == 200
    assert transport.send.call_count == 2
    assert credential.get_token.call_count == 2

    args, kwargs = credential.get_token.call_args
    assert expected_scope in args
    assert kwargs["claims"] == expected_claims

    with pytest.raises(StopIteration):
        next(tokens)
    with pytest.raises(StopIteration):
        next(responses)


def test_multiple_claims_challenges():
    """ARMChallengeAuthenticationPolicy should not attempt to handle a response having multiple claims challenges"""

    expected_header = ",".join(
        (
            'Bearer realm="", authorization_uri="https://login.microsoftonline.com/common/oauth2/authorize", client_id="00000003-0000-0000-c000-000000000000", error="insufficient_claims", claims="eyJhY2Nlc3NfdG9rZW4iOiB7ImZvbyI6ICJiYXIifX0="',
            'Bearer authorization_uri="https://login.windows-ppe.net/", error="invalid_token", error_description="User session has been revoked", claims="eyJhY2Nlc3NfdG9rZW4iOnsibmJmIjp7ImVzc2VudGlhbCI6dHJ1ZSwgInZhbHVlIjoiMTYwMzc0MjgwMCJ9fX0="',
        )
    )

    def send(request):
        return Mock(status_code=401, headers={"WWW-Authenticate": expected_header})

    transport = Mock(send=Mock(wraps=send))
    credential = FakeTokenCredential()
    policies = [ARMChallengeAuthenticationPolicy(credential, "scope")]
    pipeline = Pipeline(transport=transport, policies=policies)

    response = pipeline.run(HttpRequest("GET", "https://localhost"))

    assert transport.send.call_count == 1
    assert credential.get_token_count == 1

    # the policy should have returned the error response because it was unable to handle the challenge
    assert response.http_response.status_code == 401
    assert response.http_response.headers["WWW-Authenticate"] == expected_header



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import functools
import json

from azure.core.pipeline.transport._base import _HttpResponseBase

from azure.core.exceptions import HttpResponseError
from azure.mgmt.core.exceptions import ARMErrorFormat, TypedErrorInfo


ARMError = functools.partial(HttpResponseError, error_format=ARMErrorFormat)


def _build_response(json_body):
    class MockResponse(_HttpResponseBase):
        def __init__(self):
            super(MockResponse, self).__init__(
                request=None,
                internal_response=None,
            )
            self.status_code = 400
            self.reason = "Bad Request"
            self.content_type = "application/json"
            self._body = json_body

        def body(self):
            return self._body

    return MockResponse()


def test_arm_exception():

    message = {
        "error": {
            "code": "BadArgument",
            "message": "The provided database 'foo' has an invalid username.",
            "target": "query",
            "details": [
                {
                    "code": "301",
                    "target": "$search",
                    "message": "$search query option not supported",
                }
            ],
            "innererror": {"customKey": "customValue"},
            "additionalInfo": [{"type": "SomeErrorType", "info": {"customKey": "customValue"}}],
        }
    }
    cloud_exp = ARMError(response=_build_response(json.dumps(message).encode("utf-8")))
    assert cloud_exp.error.target == "query"
    assert cloud_exp.error.details[0].target == "$search"
    assert cloud_exp.error.innererror["customKey"] == "customValue"
    assert cloud_exp.error.additional_info[0].type == "SomeErrorType"
    assert cloud_exp.error.additional_info[0].info["customKey"] == "customValue"
    assert "customValue" in str(cloud_exp)

    message = {
        "error": {
            "code": "BadArgument",
            "message": "The provided database 'foo' has an invalid username.",
            "target": "query",
            "details": [
                {
                    "code": "301",
                    "target": "$search",
                    "message": "$search query option not supported",
                    "additionalInfo": [
                        {
                            "type": "PolicyViolation",
                            "info": {
                                "policyDefinitionDisplayName": "Allowed locations",
                                "policyAssignmentParameters": {"listOfAllowedLocations": {"value": ["westus"]}},
                            },
                        }
                    ],
                }
            ],
            "additionalInfo": [{"type": "SomeErrorType", "info": {"customKey": "customValue"}}],
        }
    }
    cloud_exp = ARMError(response=_build_response(json.dumps(message).encode("utf-8")))
    assert cloud_exp.error.target == "query"
    assert cloud_exp.error.details[0].target == "$search"
    assert cloud_exp.error.additional_info[0].type == "SomeErrorType"
    assert cloud_exp.error.additional_info[0].info["customKey"] == "customValue"
    assert cloud_exp.error.details[0].additional_info[0].type == "PolicyViolation"
    assert cloud_exp.error.details[0].additional_info[0].info["policyDefinitionDisplayName"] == "Allowed locations"
    assert (
        cloud_exp.error.details[0]
        .additional_info[0]
        .info["policyAssignmentParameters"]["listOfAllowedLocations"]["value"][0]
        == "westus"
    )
    assert "customValue" in str(cloud_exp)

    error = ARMError(response=_build_response(b"{{"))
    assert "Bad Request" in error.message

    error = ARMError(
        response=_build_response(
            b'{"error":{"code":"Conflict","message":"The maximum number of Free ServerFarms allowed in a Subscription is 10.","target":null,"details":[{"message":"The maximum number of Free ServerFarms allowed in a Subscription is 10."},{"code":"Conflict"},{"errorentity":{"code":"Conflict","message":"The maximum number of Free ServerFarms allowed in a Subscription is 10.","extendedCode":"59301","messageTemplate":"The maximum number of {0} ServerFarms allowed in a Subscription is {1}.","parameters":["Free","10"],"innerErrors":null}}],"innererror":null}}'
        )
    )
    assert error.error.code == "Conflict"

    message = json.dumps(
        {
            "error": {
                "code": "BadArgument",
                "message": "The provided database 'foo' has an invalid username.",
                "target": "query",
                "details": [
                    {
                        "code": "301",
                        "target": "$search",
                        "message": "$search query option not supported",
                    }
                ],
            }
        }
    ).encode("utf-8")
    error = ARMError(response=_build_response(message))
    assert error.error.code == "BadArgument"

    # See https://github.com/Azure/msrestazure-for-python/issues/54
    response_text = b'"{\\"error\\": {\\"code\\": \\"ResourceGroupNotFound\\", \\"message\\": \\"Resource group \'res_grp\' could not be found.\\"}}"'
    error = ARMError(response=_build_response(response_text))
    # Do not raise is plenty enough, that's a major server issue....



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
import base64
import time
from typing import Optional, Union, MutableMapping, List, Any, Sequence, TypeVar, Generic

from azure.core.credentials import AccessToken, TokenCredential
from azure.core.credentials_async import AsyncTokenCredential
from azure.core.pipeline.policies import BearerTokenCredentialPolicy, SansIOHTTPPolicy
from azure.core.pipeline import PipelineRequest, PipelineResponse
from azure.core.exceptions import ServiceRequestError
from azure.core.pipeline.transport import (
    HttpRequest as LegacyHttpRequest,
    HttpResponse as LegacyHttpResponse,
)
from azure.core.rest import HttpRequest, HttpResponse


HTTPRequestType = Union[LegacyHttpRequest, HttpRequest]
HTTPResponseType = Union[LegacyHttpResponse, HttpResponse]
TokenCredentialType = TypeVar("TokenCredentialType", bound=Union[TokenCredential, AsyncTokenCredential])


class ARMChallengeAuthenticationPolicy(BearerTokenCredentialPolicy):
    """Adds a bearer token Authorization header to requests.

    This policy internally handles Continuous Access Evaluation (CAE) challenges. When it can't complete a challenge,
    it will return the 401 (unauthorized) response from ARM.

    :param ~azure.core.credentials.TokenCredential credential: credential for authorizing requests
    :param str scopes: required authentication scopes
    """

    def on_challenge(
        self,
        request: PipelineRequest[HTTPRequestType],
        response: PipelineResponse[HTTPRequestType, HTTPResponseType],
    ) -> bool:
        """Authorize request according to an ARM authentication challenge

        :param ~azure.core.pipeline.PipelineRequest request: the request which elicited an authentication challenge
        :param ~azure.core.pipeline.PipelineResponse response: ARM's response
        :returns: a bool indicating whether the policy should send the request
        :rtype: bool
        """

        challenge = response.http_response.headers.get("WWW-Authenticate")
        if challenge:
            claims = _parse_claims_challenge(challenge)
            if claims:
                self.authorize_request(request, *self._scopes, claims=claims)
                return True

        return False


# pylint:disable=too-few-public-methods
class _AuxiliaryAuthenticationPolicyBase(Generic[TokenCredentialType]):
    """Adds auxiliary authorization token header to requests.

    :param ~azure.core.credentials.TokenCredential auxiliary_credentials: auxiliary credential for authorizing requests
    :param str scopes: required authentication scopes
    """

    def __init__(  # pylint: disable=unused-argument
        self, auxiliary_credentials: Sequence[TokenCredentialType], *scopes: str, **kwargs: Any
    ) -> None:
        self._auxiliary_credentials = auxiliary_credentials
        self._scopes = scopes
        self._aux_tokens: Optional[List[AccessToken]] = None

    @staticmethod
    def _enforce_https(request: PipelineRequest[HTTPRequestType]) -> None:
        # move 'enforce_https' from options to context, so it persists
        # across retries but isn't passed to transport implementation
        option = request.context.options.pop("enforce_https", None)

        # True is the default setting; we needn't preserve an explicit opt in to the default behavior
        if option is False:
            request.context["enforce_https"] = option

        enforce_https = request.context.get("enforce_https", True)
        if enforce_https and not request.http_request.url.lower().startswith("https"):
            raise ServiceRequestError(
                "Bearer token authentication is not permitted for non-TLS protected (non-https) URLs."
            )

    def _update_headers(self, headers: MutableMapping[str, str]) -> None:
        """Updates the x-ms-authorization-auxiliary header with the auxiliary token.

        :param dict headers: The HTTP Request headers
        """
        if self._aux_tokens:
            headers["x-ms-authorization-auxiliary"] = ", ".join(
                "Bearer {}".format(token.token) for token in self._aux_tokens
            )

    @property
    def _need_new_aux_tokens(self) -> bool:
        if not self._aux_tokens:
            return True
        for token in self._aux_tokens:
            if token.expires_on - time.time() < 300:
                return True
        return False


class AuxiliaryAuthenticationPolicy(
    _AuxiliaryAuthenticationPolicyBase[TokenCredential],
    SansIOHTTPPolicy[HTTPRequestType, HTTPResponseType],
):
    def _get_auxiliary_tokens(self, *scopes: str, **kwargs: Any) -> Optional[List[AccessToken]]:
        if self._auxiliary_credentials:
            return [cred.get_token(*scopes, **kwargs) for cred in self._auxiliary_credentials]
        return None

    def on_request(self, request: PipelineRequest[HTTPRequestType]) -> None:
        """Called before the policy sends a request.

        The base implementation authorizes the request with an auxiliary authorization token.

        :param ~azure.core.pipeline.PipelineRequest request: the request
        """
        self._enforce_https(request)

        if self._need_new_aux_tokens:
            self._aux_tokens = self._get_auxiliary_tokens(*self._scopes)

        self._update_headers(request.http_request.headers)


def _parse_claims_challenge(challenge: str) -> Optional[str]:
    """Parse the "claims" parameter from an authentication challenge

    Example challenge with claims:
        Bearer authorization_uri="https://login.windows-ppe.net/", error="invalid_token",
        error_description="User session has been revoked",
        claims="eyJhY2Nlc3NfdG9rZW4iOnsibmJmIjp7ImVzc2VudGlhbCI6dHJ1ZSwgInZhbHVlIjoiMTYwMzc0MjgwMCJ9fX0="

    :param str challenge: The authentication challenge
    :return: the challenge's "claims" parameter or None, if it doesn't contain that parameter
    """
    encoded_claims = None
    for parameter in challenge.split(","):
        if "claims=" in parameter:
            if encoded_claims:
                # multiple claims challenges, e.g. for cross-tenant auth, would require special handling
                return None
            encoded_claims = parameter[parameter.index("=") + 1 :].strip(" \"'")

    if not encoded_claims:
        return None

    padding_needed = -len(encoded_claims) % 4
    try:
        decoded_claims = base64.urlsafe_b64decode(encoded_claims + "=" * padding_needed).decode()
        return decoded_claims
    except Exception:  # pylint:disable=broad-except
        return None
